import gym
from gym import spaces
from gym.utils import seeding
import numpy as np
import copy
import timemeasurer
import LatencyMatrix
import AWS
import AppWorkload
import Tracedata
import LSTMdata


map = LatencyMatrix.latencymap
clouds = AWS.datacenter
centers = 18
workload = AppWorkload.WikiWorkload
requestdist = Tracedata.distributions
prediction = LSTMdata.workload
multistep = 5
capacityunit = 100
rho = 50
maxPeriods = 480
bound = 150


# Capacity-based weighted Round-Robin
def getschedules(plan, requests):
    total_cpu = sum(plan)
    schedules = []
    for u in range(82):
        schedule = []
        for c in range(centers):
            schedule.append(requests[u] * plan[c] / total_cpu)
        schedules.append(schedule)
    return schedules

# Calculate the aggregated workloads of application instances
def getinstanceworkload(schedules):
    workload = [0] * centers
    for c in range(centers):
        for schedule in schedules:
            workload[c] += schedule[c]
    return workload

# Estimate the average CPU utilization
def geturate(plan, workload):
    totalcpu = 0
    totalworkload = 0
    for c in range(centers):
        totalcpu += plan[c]
        totalworkload += workload[c]
    return totalworkload / (totalcpu * capacityunit)

# Calculate the average network latency
def getanl(schedules):
    workload = getinstanceworkload(schedules)
    if sum(workload) > 0:
        latency = 0
        for c in range(centers):
            for u in range(82):
                latency += map[u][clouds[c][0]] * schedules[u][c]
        ant = latency / sum(workload)
    else:
        ant = 0
    return ant

#Calculate the cost
def getcost(plan):
    cost = 0
    for c in range(centers):
        if plan[c] > 0:
            cost += plan[c] * clouds[c][2][0] + clouds[c][2][1]
    return cost / 60 * 3

#Decide the data center with the highest scale-up benefit
def getaddcenter(plan, requests):
    oldcost = getcost(plan)
    schedules = getschedules(plan, requests)
    oldtime = getanl(schedules)
    des = -1
    best = 0
    for c in range(centers):
        if plan[c] < 20:
            newplan = copy.deepcopy(plan)
            newplan[c] += 1
            newcost = getcost(newplan)
            schedules = getschedules(newplan, requests)
            newtime = getanl(schedules)
            benefit = (oldtime - newtime) / (newcost - oldcost)
            if benefit > best:
                des = c
                best = benefit
    return des

#Decide the data center with the highest scale-down benefit
def getreducecenter(plan, requests):
    oldcost = getcost(plan)
    schedules = getschedules(plan, requests)
    oldtime = getanl(schedules)
    des = -1
    best = 0
    for c in range(centers):
        if plan[c] > 0:
            newplan = copy.deepcopy(plan)
            newplan[c] -= 1
            newcost = getcost(newplan)
            schedules = getschedules(newplan, requests)
            newtime = getanl(schedules)
            benefit = (oldcost - newcost) / (newtime - oldtime)
            if benefit > best:
                des = c
                best = benefit
    return des

#Scale-up CPUs
def addcpu(plan, requests):
    newplan = copy.deepcopy(plan)
    des = getaddcenter(newplan, requests)
    if newplan[des] < 20:
        newplan[des] += 1
    while geturate(newplan, requests) >= 1:
        des = getaddcenter(newplan, requests)
        if newplan[des] < 20:
            newplan[des] += 1
    return newplan

# Maintain CPUs
def keepcpu(plan, requests):
    newplan = copy.deepcopy(plan)
    des = getaddcenter(newplan, requests)
    if newplan[des] < 20:
        newplan[des] += 1
        des = getreducecenter(newplan, requests)
        newplan[des] -= 1
    return newplan

# Scale-down CPUs
def reducecpu(plan, requests):
    newplan = copy.deepcopy(plan)
    total = sum(requests)
    capacity = sum(newplan) * capacityunit
    # while capacity - unit >= total:
    if capacity - capacityunit > total:
        #print("do", capacity, total)
        des = getreducecenter(newplan, requests)
        newplan[des] -= 1
        # capacity = sum(newplan) * unit
    return newplan

def gettrend(periodNum):
    trend = []
    for step in range(multistep):
        trend.append(prediction[periodNum + step + 1] - prediction[periodNum + step])
    return trend


class DeepScale(gym.Env):

    def __init__(self):
        low = np.array([0.0] * centers + [0] + [-10.0] * multistep, dtype=np.float32)
        high = np.array([20.0] * centers + [1.0] + [10.0] * multistep, dtype=np.float32)
        self.observation_space = spaces.Box(low, high, dtype=np.float32)

        self.seed()
        self.reset()

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def getworkload(self, periodNum):
        return workload[periodNum + 1]

    def step(self, action, periodNum):
        #Scale-up
        if action == 0:
            newplan = addcpu(self.plan, self.user)
        #Maintain
        elif action == 1:
            newplan = keepcpu(self.plan, self.user)
        #Scale-down
        else:
            newplan = reducecpu(self.plan, self.user)
        newcost = getcost(newplan)
        newtime = timemeasurer.getart()
        self.done[periodNum] = True
        self.cost[periodNum] = newcost
        self.time[periodNum] = newtime
        self.reward[periodNum] = 0 - newcost - max(0, rho * (newtime - bound))

        if periodNum + 1 < maxPeriods:
            self.done[periodNum + 1] = False
            self.load = workload[periodNum + 1]
            self.trend = gettrend(periodNum)
            self.user = requestdist[periodNum + 1]
            self.plan = newplan
            self.utilization = geturate(self.plan, self.load)

        return self.getObs(), self.reward, self.cost, self.ant, self.urate, self.done

    def getObs(self):
        state = copy.deepcopy(self.plan)
        state += [self.utilization]
        state += self.trend
        return state

    # Reset environment
    def reset(self):
        self.load = workload[0]
        self.trend = gettrend(0)
        self.user = requestdist[0]
        self.plan = [1] * 8 + [0] * 10
        self.utilization = workload[0] / (sum(self.plan) * capacityunit)
        self.cost = {}
        self.time = {}
        self.reward = {}
        self.done = {}
        return self.getObs()